"""
Dashboard registry.
"""


class Registry(object):
    """
    Registry for application dashboards.
    """

    registry = {}

    def register(cls, klass, app_name):
        from jnt_admin_tools.dashboard.dashboards import Dashboard

        if not issubclass(klass, Dashboard):
            raise ValueError("%s is not an instance of Dashboard" % klass)
        if app_name in cls.registry:
            raise ValueError(
                'A dashboard has already been registered for app "%s"'
                % app_name
            )
        cls.registry[app_name] = klass

    register = classmethod(register)


def register(cls, *args, **kwargs):
    """
    Register a custom dashboard into the global registry.
    """
    Registry.register(cls, *args, **kwargs)


def autodiscover(blacklist=[]):
    """
    Automagically discover custom dashboards and menus for installed apps.
    Optionally you can pass a ``blacklist`` of apps that you don't want to
    provide their own app index dashboard.
    """
    import imp
    from django.conf import settings

    try:
        from importlib import import_module
    except ImportError:
        # Django < 1.9 and Python < 2.7
        from django.utils.importlib import import_module

    blacklist.append("jnt_admin_tools.dashboard")
    blacklist.append("jnt_admin_tools.menu")
    blacklist.append("jnt_admin_tools.theming")

    for app in settings.INSTALLED_APPS:
        # skip blacklisted apps
        if app in blacklist:
            continue

        # try to import the app
        try:
            app_path = import_module(app).__path__
        except AttributeError:
            continue

        # try to find a app.dashboard module
        try:
            imp.find_module("dashboard", app_path)
        except ImportError:
            continue

        # looks like we found it so import it !
        import_module("%s.dashboard" % app)



from django import forms
from django.contrib.admin.options import InlineModelAdmin
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext_lazy as _
from jnt_admin_tools.db.fields import GenericForeignKey
from jnt_admin_tools.services.objects import copy_func
from jnt_admin_tools.widgets.autocomplete import ContentTypeAutocompleteSelect


class BaseGenericForeignKeyMixin:
    def get_fields(self, request, obj=None):
        fields = super().get_fields(request, obj=obj)
        if fields:
            fields = self._update_gfk_fields(fields)

        return fields

    def _update_gfk_fields(self, fields):
        gfk_fields = self._get_generic_foreign_keys(self.model)

        if not gfk_fields:
            return fields

        gfk_map = {f.name: f for f in gfk_fields}

        updated = []
        for field in fields:
            gfk = gfk_map.get(field)

            if not (gfk and isinstance(self, InlineModelAdmin)):
                updated = self._append_value(updated, field)

            if gfk:
                updated = self._append_value(updated, gfk.ct_field)
                updated = self._append_value(updated, gfk.fk_field)

        return tuple(updated)

    def _get_generic_foreign_keys(self, model):
        return [
            f
            for f in model._meta.get_fields()
            if isinstance(f, GenericForeignKey)
        ]

    def _append_value(self, fields, value):
        if value not in fields:
            fields.append(value)
        return fields

    def _set_validators_gfk(self, form):
        for field in self._get_generic_foreign_keys(self.model):
            _copy_func = copy_func(self._clean_formfield)
            _copy_func.__defaults__ = (field,)

            setattr(form, f"clean_{field.name}", _copy_func)

    def _clean_formfield(self, gfk_field):
        ct_field = self.cleaned_data.get(gfk_field.ct_field)
        fk_field = self.cleaned_data.get(gfk_field.fk_field)

        if ct_field and fk_field:
            return
        elif not ct_field and not fk_field:
            return

        msg = _("Both values must be filled or cleared")
        self.add_error(gfk_field.ct_field, msg)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        generic_foreign_keys = self._get_generic_foreign_keys(self.model)

        is_gfk, gfk = self._is_generic_foreign_key(
            db_field.name, generic_foreign_keys
        )

        form_field = super().formfield_for_foreignkey(
            db_field, request, **kwargs
        )

        if not is_gfk:
            return form_field

        attrs = self._get_content_type_autocomplete_attrs(gfk)
        db = kwargs.get("using")
        form_field.widget = ContentTypeAutocompleteSelect(
            db_field.remote_field, self.admin_site, using=db, attrs=attrs
        )

        return form_field

    def _get_gfk_ids_related_models(self, gfk):
        related_models = gfk.get_related_models()

        content_types = ContentType.objects.get_for_models(*related_models)

        return [x.id for x in content_types.values()]

    def _set_declared_gfk_to_form(self, form):
        for field in self._get_generic_foreign_keys(self.model):
            form_char_field = forms.CharField(
                widget=forms.HiddenInput(),
                required=False,
            )

            form.declared_fields[field.name] = form_char_field

    def _is_generic_foreign_key(self, field_name, generic_foreign_keys):
        for gfk in generic_foreign_keys:
            if field_name in (gfk.name, gfk.ct_field, gfk.fk_field):
                return True, gfk
        return False, None

    def _get_content_type_autocomplete_attrs(self, gfk):
        return {
            "class": "generic-foreign-key-field",
            "data-ct--field": gfk.ct_field,
            "data-fk--field": gfk.fk_field,
            "data-gf--name": gfk.name,
            "data-gfk--models": ",".join(
                str(x) for x in self._get_gfk_ids_related_models(gfk)
            ),
        }



from django.contrib.admin.helpers import AdminReadonlyField
from django.core.exceptions import FieldDoesNotExist

from jnt_admin_tools.services.object_links import (
    add_object_link_present_function,
)
from jnt_admin_tools.services.readonly_form_field import (
    get_present_admin_readonly_field,
)

# Patching AdminReadonlyField
orig_contents = AdminReadonlyField.contents


def contents(self):
    return get_present_admin_readonly_field(self) or orig_contents(self)


AdminReadonlyField.contents = contents


class AdminClickableLinksMixin:
    def get_list_display(self, request):
        list_display = super().get_list_display(request)
        return self._updating_list_display(list_display, request)

    def _updating_list_display(self, list_display, request):
        if len(list_display) < 2:
            return list_display

        list_display_links = self.get_list_display_links(request, list_display)

        fields = []
        for field_name in list_display:
            is_relation = self._is_relation(
                self.model,
                field_name,
                list_display_links,
            )

            if not is_relation:
                fields.append(field_name)
                continue

            fields.append(add_object_link_present_function(self, field_name))
        return fields

    def _is_relation(self, model, field_name, list_display_links):
        field_name = field_name.split("__")[0]
        if not hasattr(model, field_name) or field_name in list_display_links:
            return False

        try:
            field = model._meta.get_field(field_name)
        except FieldDoesNotExist:
            return False

        return any(
            (
                field.many_to_one,
                field.one_to_one,
                field.many_to_many,
                field.one_to_many,
            ),
        )



from django.apps import AppConfig


class AdminToolsConfig(AppConfig):
    name = "jnt_admin_tools"

    def ready(self):
        super().ready()
        # load jnt_admin_tools checks
        from . import checks  # noqa: F401



from jnt_admin_tools.services.urls import admin_autocomplete_url
from django.apps import apps
from django.contrib.contenttypes.fields import (
    GenericForeignKey as BaseGenericForeignKey,
)


class GenericForeignKey(BaseGenericForeignKey):
    def __init__(self, *args, **kwargs):
        self.related_models = kwargs.pop("related_models", None)
        super().__init__(*args, **kwargs)

    def get_related_models(self):
        if self.related_models is None:
            self.related_models = [
                m for m in apps.get_models() if self.has_autocomplete_url(m)
            ]
        elif callable(self.related_models):
            self.related_models = list(self.related_models())
        return [
            apps.get_model(*m.split(".")) if isinstance(m, str) else m
            for m in self.related_models
        ]

    def has_autocomplete_url(self, model):
        return bool(admin_autocomplete_url(model))



from django import forms
from django.contrib.auth.models import Permission
from django.db.models import QuerySet

from jnt_admin_tools.widgets import (
    PermissionSelectMultipleReadonlyWidget,
    PermissionSelectMultipleWidget,
)


class PermissionSelectMultipleField(forms.ModelMultipleChoiceField):
    """
    A form field for displaying all permissions in system.

    Usage::

        from jnt_admin_tools.fields import PermissionSelectMultipleField
        from django import forms
        from django.contrib.auth.models import Group

        class GroupAdminForm(forms.ModelForm):
            permissions = PermissionSelectMultipleField(required=False)

        class Meta:
            model = Group
            fields = '__all__'

    .. image:: images/widgets/permission_select_multiple_widget.png
    """

    widget = PermissionSelectMultipleWidget
    readonly_widget = PermissionSelectMultipleReadonlyWidget

    def __init__(self, queryset: QuerySet = None, *args, **kwargs) -> None:
        """Initialize."""
        if queryset is None:
            queryset = Permission.objects.all()

        super().__init__(queryset, *args, **kwargs)



from jnt_admin_tools.mixins.base_generic_foreign_key import (
    BaseGenericForeignKeyMixin,
)


class GenericForeignKeyAdminMixin(BaseGenericForeignKeyMixin):
    def get_fieldsets(self, request, obj=None):
        fieldsets = super().get_fieldsets(request, obj=obj)

        if fieldsets:
            fieldsets = self._update_gfk_fieldsets(fieldsets)

        return fieldsets

    def get_form(self, request, obj=None, change=False, **kwargs):
        self._set_declared_gfk_to_form(self.form)
        form = super().get_form(request, obj=obj, change=change, **kwargs)
        self._set_present_gfk(form, obj)
        self._set_validators_gfk(form)

        return form

    def _set_present_gfk(self, form, obj):
        for field in self._get_generic_foreign_keys(self.model):
            if field.fk_field in form.base_fields:
                form.base_fields[field.fk_field].widget.attrs[
                    "data-present"
                ] = str(getattr(obj, field.name, ""))

    def get_inline_formsets(
        self, request, formsets, inline_instances, obj=None
    ):
        self.set_present_for_inlines(request, formsets, inline_instances, obj)

        return super().get_inline_formsets(
            request, formsets, inline_instances, obj=obj
        )

    def set_present_for_inlines(
        self, request, formsets, inline_instances, obj
    ):
        for formset, inline in zip(formsets, inline_instances):
            generic_foreign_keys = self._get_generic_foreign_keys(
                formset.model
            )

            if not generic_foreign_keys:
                continue

            inline_fields = inline.get_fields(request, obj)

            for form in formset.forms:
                for gfk in generic_foreign_keys:
                    if gfk.ct_field not in inline_fields:
                        continue

                    form.fields[gfk.fk_field].widget.attrs[
                        "data-present"
                    ] = str(getattr(form.instance, gfk.name, ""))

    def _update_gfk_fieldsets(self, fieldsets):
        gfk_fields = self._get_generic_foreign_keys(self.model)

        if not gfk_fields:
            return fieldsets

        for fieldset in fieldsets:
            fields = fieldset[1].get("fields")
            if fields:
                fieldset[1]["fields"] = self._update_gfk_fields(fields)
        return fieldsets
